from pathlib import Path

import numpy as np
import pytest
import qoi

# Create an empty 3-channel image
RGB = np.random.randint(low=0, high=255, size=(224, 244, 3)).astype(np.uint8)
RGBA = np.random.randint(low=0, high=255, size=(224, 244, 4)).astype(np.uint8)


@pytest.mark.parametrize("is_path", [True, False])
@pytest.mark.parametrize("is_rgba", [True, False])
@pytest.mark.parametrize("colorspace", [qoi.QOIColorSpace.SRGB, qoi.QOIColorSpace.LINEAR, None])
def test_write_read(tmp_path: Path, is_path: bool, colorspace: qoi.QOIColorSpace, is_rgba: bool):
    tmp_path = str(tmp_path) + "tmp.qoi"
    img = RGBA if is_rgba else RGB
    if is_path:
        tmp_path = Path(tmp_path)
    if colorspace is None:
        bytes_written = qoi.write(tmp_path, img)
    else:
        bytes_written = qoi.write(tmp_path, img, colorspace)
    assert bytes_written > 0
    img_read = qoi.read(tmp_path)
    assert np.array_equal(img, img_read)


@pytest.mark.parametrize("is_rgba", [True, False])
@pytest.mark.parametrize("colorspace", [qoi.QOIColorSpace.SRGB, qoi.QOIColorSpace.LINEAR, None])
def test_encode_decode(colorspace: qoi.QOIColorSpace, is_rgba: bool):
    img = RGBA if is_rgba else RGB
    if colorspace is None:
        bites = qoi.encode(img)
    else:
        bites = qoi.encode(img, colorspace)
    img_decoded = qoi.decode(bites)
    assert np.array_equal(img, img_decoded)


@pytest.mark.parametrize("is_rgba", [True, False])
@pytest.mark.parametrize("colorspace", [qoi.QOIColorSpace.SRGB, qoi.QOIColorSpace.LINEAR, None])
def test_encode_decode_with_colorspace(colorspace: qoi.QOIColorSpace, is_rgba: bool):
    img = RGBA if is_rgba else RGB
    if colorspace is None:
        bites = qoi.encode(img)
    else:
        bites = qoi.encode(img, colorspace)
    data = bytearray(1)
    img_decoded = qoi.decode(bites, colorspace=data)
    if colorspace is not None:
        assert data[0] == colorspace.value
    else:
        assert data[0] == qoi.QOIColorSpace.SRGB.value
    assert np.array_equal(img, img_decoded)


def test_version():
    assert qoi.__version__ is not None


def test_writes_without_extension(tmp_path):
    qoi.write(str(tmp_path) + "tmp", RGB, qoi.QOIColorSpace.SRGB)


def test_benchmark():
    from qoi.benchmark import benchmark

    benchmark(implementations=["qoi"])



cimport qoi.qoi as qoi
import numpy as np
cimport numpy as np
from cpython cimport PyObject, Py_INCREF
from cpython.mem cimport PyMem_Free
import enum
from pathlib import Path

np.import_array()

ctypedef np.uint8_t DTYPE_t

class QOIColorSpace(enum.Enum):
    SRGB = qoi.QOI_SRGB
    LINEAR = qoi.QOI_LINEAR

cdef class PixelWrapper:
    cdef void* pixels

    cdef np.ndarray[DTYPE_t, ndim=3] as_ndarray(self, int height, int width, int channels, char * pixels):
        cdef np.npy_intp shape[3]
        cdef np.ndarray[DTYPE_t, ndim=3] ndarray
        self.pixels = pixels
        shape[:] = (height, width, channels)
        ndarray = np.PyArray_SimpleNewFromData(3, shape, np.NPY_UINT8, self.pixels)
        ndarray.base = <PyObject*> self
        Py_INCREF(self)
        return ndarray

    def __dealloc__(self):
        PyMem_Free(self.pixels)

cpdef int write(filename, unsigned char[:,:,::1] rgb, colorspace: QOIColorSpace = QOIColorSpace.SRGB) except? -1:
    cdef bytes filename_bytes 
    cdef char* _filename
    cdef qoi.qoi_desc desc
    cdef int ret

    if not isinstance(colorspace, QOIColorSpace):
        raise ValueError("colorspace should be an instance of QOIColorSpace")
    if not isinstance(filename, (str, Path)):
        raise ValueError("filename should be a str or Path")

    filename_bytes = str(filename).encode('utf8')
    _filename = filename_bytes
    
    desc.height = <unsigned int>rgb.shape[0]
    desc.width = <unsigned int>rgb.shape[1]
    desc.channels = <unsigned char>rgb.shape[2]
    desc.colorspace = colorspace.value
    
    # if not rgb.flags['C_CONTIGUOUS']:
        # Makes a contiguous copy of the numpy array so we can process bytes directly:
        # rgb = np.ascontiguousarray(rgb)
    
    cdef int bytes_written
    with nogil:
        bytes_written = qoi.qoi_write(_filename, &rgb[0][0][0], &desc)
    if bytes_written == 0:
        raise RuntimeError("Failed to write!")
    return bytes_written

cpdef np.ndarray[DTYPE_t, ndim=3] read(filename, int channels = 0, unsigned char[::1] colorspace = bytearray(1)):
    # TODO: how to return desc.colorspace? A: How about return a tuple of ndarray and a wrapper around struct qoi_desc? or we can add another param like char[:] to simulate pointer
    cdef bytes filename_bytes
    cdef char* _filename
    cdef qoi.qoi_desc desc
    cdef int ret
    cdef char* pixels

    if not isinstance(filename, (str, Path)):
        raise ValueError("filename should be a str or Path")

    filename_bytes = str(filename).encode('utf8')
    _filename = filename_bytes
    
    with nogil:
        pixels = <char *>qoi.qoi_read(_filename, &desc, channels)
    if pixels is NULL:
        raise RuntimeError("Failed to read!")
    try:
        colorspace[0] = desc.colorspace
        return PixelWrapper().as_ndarray(desc.height, desc.width, desc.channels if channels == 0 else channels, pixels)
    except:
        if pixels is not NULL:
            PyMem_Free(pixels)

cpdef bytes encode(unsigned char[:,:,::1] rgb, colorspace: QOIColorSpace = QOIColorSpace.SRGB):
    cdef qoi.qoi_desc desc
    cdef int ret, size
    cdef char * encoded

    if not isinstance(colorspace, QOIColorSpace):
        raise ValueError("colorspace should be an instance of QOIColorSpace")

    desc.height = <unsigned int>rgb.shape[0]
    desc.width = <unsigned int>rgb.shape[1]
    desc.channels = <unsigned char>rgb.shape[2]
    desc.colorspace = colorspace.value

    # if not rgb.flags['C_CONTIGUOUS']:
    #     rgb = np.ascontiguousarray(rgb) # makes a contiguous copy of the numpy array so we can read memory directly
    with nogil:
        encoded = <char *>qoi.qoi_encode(&rgb[0][0][0], &desc, &size)
    if encoded is NULL or size <= 0:
        raise RuntimeError("Failed to encode!")
    try:
        # TODO: does this create a copy? A: Yes, this equivalent to PyBytes_FromStringAndSize
        return encoded[:size] # :size is important here - tells cython about size, and handles null bytes
    finally:
        PyMem_Free(encoded)

cpdef np.ndarray[DTYPE_t, ndim=3] decode(const unsigned char[::1] data, int channels = 0, unsigned char[::1] colorspace = bytearray(1)):
    # TODO: what to do about desc.colorspace? A: How about return a tuple of ndarray and a wrapper around struct qoi_desc? or we can add another param like char[:] to simulate pointer
    cdef qoi.qoi_desc desc
    cdef int ret
    cdef char * pixels
    with nogil:
        pixels = <char *>qoi.qoi_decode(&data[0], <int>data.shape[0], &desc, channels)
    if pixels is NULL:
        raise RuntimeError("Failed to decode!")
    try:
        colorspace[0] = desc.colorspace
        return PixelWrapper().as_ndarray(desc.height, desc.width, desc.channels if channels == 0 else channels, pixels)
    except:
        if pixels is not NULL:
            PyMem_Free(pixels)



import os

import numpy as np
import setuptools_scm  # noqa: F401 - to avoid version = 0.0.0 errors if built without setuptools_scm installed
from setuptools import Extension, setup

USE_CYTHON = os.getenv("USE_CYTHON") == "1"

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None
    if USE_CYTHON:
        raise RuntimeError("You've set USE_CYTHON=1 but don't have Cython installed!")


# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


extensions = [
    Extension(
        "qoi.qoi",
        sources=["src/qoi/qoi.pyx", "src/qoi/implementation.c"],
        define_macros=[("QOI_MALLOC", "PyMem_Malloc"), ("QOI_FREE", "PyMem_Free")],
    )
]


if USE_CYTHON:
    compiler_directives = {
        "language_level": 3,
        "embedsignature": True,
        "boundscheck": False,
        "wraparound": False,
        "cdivision": True,
    }
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)

with open("requirements.txt") as f:
    install_requires = f.read().strip().split("\n")

with open("requirements-dev.txt") as f:
    dev_requires = f.read().strip().split("\n")

setup(
    ext_modules=extensions,
    install_requires=install_requires,
    extras_require={
        "dev": dev_requires,
    },
    include_dirs=[np.get_include()],
)



from concurrent.futures import ThreadPoolExecutor, wait

import numpy as np
import qoi

RGB = np.random.randint(low=0, high=255, size=(224, 244, 3)).astype(np.uint8)


def worker():
    bites = bytearray(qoi.encode(RGB))
    decoded = qoi.decode(bites)
    assert np.array_equal(RGB, decoded)


def test_multithreaded():
    """
    Note that this doesn't really test the performance, but it at least validates that it runs multithreaded.
    """
    with ThreadPoolExecutor(8) as pool:
        futures = [pool.submit(worker) for _ in range(100)]
        wait(futures)
